@@ -48,7 +48,7 @@
 
 /* add/remove Virtual IP addresses */
 static int
-vrrp_handle_ipaddress(vrrp_rt * vrrp, int cmd, int type)
+vrrp_handle_ipaddress(vrrp_t * vrrp, int cmd, int type)
 {
 	if (debug & 32)
 		log_message(LOG_INFO, "VRRP_Instance(%s) %s protocol %s", vrrp->iname,
@@ -60,46 +60,46 @@ vrrp_handle_ipaddress(vrrp_rt * vrrp, int cmd, int type)
 
 /* add/remove Virtual routes */
 static int
-vrrp_handle_iproutes(vrrp_rt * vrrp, int cmd)
+vrrp_handle_iproutes(vrrp_t * vrrp, int cmd)
 {
 	if (debug & 32)
 		log_message(LOG_INFO, "VRRP_Instance(%s) %s protocol Virtual Routes",
 		       vrrp->iname,
 		       (cmd == IPROUTE_ADD) ? "setting" : "removing");
-	netlink_rtlist_ipv4(vrrp->vroutes, cmd);
+	netlink_rtlist(vrrp->vroutes, cmd);
 	return 1;
 }
 
 /* IP header length */
 static int
-vrrp_iphdr_len(vrrp_rt * vrrp)
+vrrp_iphdr_len(vrrp_t * vrrp)
 {
-	return sizeof (struct iphdr);
+	return sizeof(struct iphdr);
 }
 
 /* IPSEC AH header length */
 int
 vrrp_ipsecah_len(void)
 {
-	return sizeof (ipsec_ah);
+	return sizeof(ipsec_ah_t);
 }
 
 /* VRRP header length */
 static int
-vrrp_hd_len(vrrp_rt * vrrp)
+vrrp_hd_len(vrrp_t * vrrp)
 {
-	int len = sizeof (vrrp_pkt);
+	int len = sizeof(vrrphdr_t);
 	if (vrrp->family == AF_INET)
 		len += VRRP_AUTH_LEN + ((!LIST_ISEMPTY(vrrp->vip)) ? LIST_SIZE(vrrp->vip) * sizeof (uint32_t) : 0);
         return len;
 }
 
 /* VRRP header pointer from buffer */
-vrrp_pkt *
+vrrphdr_t *
 vrrp_get_header(sa_family_t family, char *buf, int *proto, uint32_t *saddr)
 {
 	struct iphdr *iph;
-	vrrp_pkt *hd = NULL;
+	vrrphdr_t *hd = NULL;
 
 	if (family == AF_INET) {
 		iph = (struct iphdr *) buf;
@@ -109,18 +109,18 @@ vrrp_get_header(sa_family_t family, char *buf, int *proto, uint32_t *saddr)
 		switch (iph->protocol) {
 		case IPPROTO_IPSEC_AH:
 			*proto = IPPROTO_IPSEC_AH;
-			hd = (vrrp_pkt *) ((char *) iph + (iph->ihl << 2) +
+			hd = (vrrphdr_t *) ((char *) iph + (iph->ihl << 2) +
 					   vrrp_ipsecah_len());
 			break;
 		case IPPROTO_VRRP:
 			*proto = IPPROTO_VRRP;
-			hd = (vrrp_pkt *) ((char *) iph + (iph->ihl << 2));
+			hd = (vrrphdr_t *) ((char *) iph + (iph->ihl << 2));
 			break;
 		}
 	} else if (family == AF_INET6) {
 		*proto = IPPROTO_VRRP;
 		*saddr = 0;
-		hd = (vrrp_pkt *) buf;
+		hd = (vrrphdr_t *) buf;
 	}
 
 	return hd;
@@ -131,10 +131,10 @@ vrrp_get_header(sa_family_t family, char *buf, int *proto, uint32_t *saddr)
  * return 0 for a valid pkt, != 0 otherwise.
  */
 static int
-vrrp_in_chk_ipsecah(vrrp_rt * vrrp, char *buffer)
+vrrp_in_chk_ipsecah(vrrp_t * vrrp, char *buffer)
 {
 	struct iphdr *ip = (struct iphdr *) (buffer);
-	ipsec_ah *ah = (ipsec_ah *) ((char *) ip + (ip->ihl << 2));
+	ipsec_ah_t *ah = (ipsec_ah_t *) ((char *) ip + (ip->ihl << 2));
 	unsigned char *digest;
 	uint32_t backup_auth_data[3];
 
@@ -166,7 +166,7 @@ vrrp_in_chk_ipsecah(vrrp_rt * vrrp, char *buffer)
 	 * then compute a ICV to compare with the one present in AH pkt.
 	 * alloc a temp memory space to stock the ip mutable fields
 	 */
-	digest = (unsigned char *) MALLOC(16 * sizeof (unsigned char *));
+	digest = (unsigned char *) MALLOC(16); /*MD5_DIGEST_LENGTH */
 
 	/* zero the ip mutable fields */
 	ip->tos = 0;
@@ -194,7 +194,7 @@ vrrp_in_chk_ipsecah(vrrp_rt * vrrp, char *buffer)
 
 /* check if ipaddr is present in VIP buffer */
 static int
-vrrp_in_chk_vips(vrrp_rt * vrrp, ip_address *ipaddress, unsigned char *buffer)
+vrrp_in_chk_vips(vrrp_t * vrrp, ip_address_t *ipaddress, unsigned char *buffer)
 {
 	int i;
 	uint32_t ipbuf;
@@ -220,14 +220,14 @@ vrrp_in_chk_vips(vrrp_rt * vrrp, ip_address *ipaddress, unsigned char *buffer)
  * return 0 if the pkt is valid, != 0 otherwise.
  */
 static int
-vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
+vrrp_in_chk(vrrp_t * vrrp, char *buffer)
 {
 	struct iphdr *ip;
-	int ihl, vrrp_pkt_len;
-	ipsec_ah *ah;
-	vrrp_pkt *hd;
+	int ihl, vrrphdr_len;
+	ipsec_ah_t *ah;
+	vrrphdr_t *hd;
 	unsigned char *vips;
-	ip_address *ipaddress;
+	ip_address_t *ipaddress;
 	element e;
 
 	/* IPv4 related */
@@ -237,17 +237,17 @@ vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
 		ihl = ip->ihl << 2;
 
 		if (vrrp->auth_type == VRRP_AUTH_AH) {
-			ah = (ipsec_ah *) (buffer + ihl);
-			hd = (vrrp_pkt *) (ah + vrrp_ipsecah_len());
+			ah = (ipsec_ah_t *) (buffer + ihl);
+			hd = (vrrphdr_t *) ((char *) ah + vrrp_ipsecah_len());
 		} else {
-			hd = (vrrp_pkt *) (buffer + ihl);
+			hd = (vrrphdr_t *) (buffer + ihl);
 		}
 	
 		/* pointer to vrrp vips pkt zone */
-		vips = (unsigned char *) ((char *) hd + sizeof (vrrp_pkt));
+		vips = (unsigned char *) ((char *) hd + sizeof(vrrphdr_t));
 	
 		/* MUST verify that the IP TTL is 255 */
-		if (ip->ttl != VRRP_IP_TTL) {
+		if (LIST_ISEMPTY(vrrp->unicast_peer) && ip->ttl != VRRP_IP_TTL) {
 			log_message(LOG_INFO, "invalid ttl. %d and expect %d", ip->ttl,
 			       VRRP_IP_TTL);
 			return VRRP_PACKET_KO;
@@ -257,10 +257,10 @@ vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
 		 * MUST verify that the received packet length is greater than or
 		 * equal to the VRRP header
 		 */
-		if ((ntohs(ip->tot_len) - ihl) <= sizeof (vrrp_pkt)) {
+		if ((ntohs(ip->tot_len) - ihl) <= sizeof(vrrphdr_t)) {
 			log_message(LOG_INFO,
 			       "ip payload too short. %d and expect at least %d",
-			       ntohs(ip->tot_len) - ihl, sizeof (vrrp_pkt));
+			       ntohs(ip->tot_len) - ihl, sizeof(vrrphdr_t));
 			return VRRP_PACKET_KO;
 		}
 
@@ -304,12 +304,12 @@ vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
 			return (vrrp_in_chk_ipsecah(vrrp, buffer));
 
 		/* Set expected vrrp packet lenght */
-		vrrp_pkt_len = sizeof(vrrp_pkt) + VRRP_AUTH_LEN + hd->naddr * sizeof(uint32_t);
+		vrrphdr_len = sizeof(vrrphdr_t) + VRRP_AUTH_LEN + hd->naddr * sizeof(uint32_t);
 
 	} else if (vrrp->family == AF_INET6) { /* IPv6 related */
 
-		hd = (vrrp_pkt *) buffer;
-		vrrp_pkt_len = sizeof(vrrp_pkt);
+		hd = (vrrphdr_t *) buffer;
+		vrrphdr_len = sizeof(vrrphdr_t);
 
 	} else {
 		return VRRP_PACKET_KO;
@@ -323,7 +323,7 @@ vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
 	}
 
 	/* MUST verify the VRRP checksum */
-	if (in_csum((u_short *) hd, vrrp_pkt_len, 0)) {
+	if (in_csum((u_short *) hd, vrrphdr_len, 0)) {
 		log_message(LOG_INFO, "Invalid vrrp checksum");
 		return VRRP_PACKET_KO;
 	}
@@ -338,7 +338,7 @@ vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
 		return VRRP_PACKET_KO;
 	}
 
-	/* MUST verify that the VRID is valid on the receiving interface */
+	/* MUST verify that the VRID is valid on the receiving interface_t */
 	if (vrrp->vrid != hd->vrid) {
 		log_message(LOG_INFO,
 		       "received VRID mismatch. Received %d, Expected %d",
@@ -367,7 +367,7 @@ vrrp_in_chk(vrrp_rt * vrrp, char *buffer)
 
 /* build IP header */
 static void
-vrrp_build_ip(vrrp_rt * vrrp, char *buffer, int buflen)
+vrrp_build_ip4(vrrp_t * vrrp, char *buffer, int buflen, uint32_t dst)
 {
 	struct iphdr *ip = (struct iphdr *) (buffer);
 
@@ -385,10 +385,9 @@ vrrp_build_ip(vrrp_rt * vrrp, char *buffer, int buflen)
 	ip->ttl = VRRP_IP_TTL;
 
 	/* fill protocol type --rfc2402.2 */
-	ip->protocol =
-	    (vrrp->auth_type == VRRP_AUTH_AH) ? IPPROTO_IPSEC_AH : IPPROTO_VRRP;
+	ip->protocol = (vrrp->auth_type == VRRP_AUTH_AH) ? IPPROTO_IPSEC_AH : IPPROTO_VRRP;
 	ip->saddr = VRRP_PKT_SADDR(vrrp);
-	ip->daddr = htonl(INADDR_VRRP_GROUP);
+	ip->daddr = dst;
 
 	/* checksum must be done last */
 	ip->check = in_csum((u_short *) ip, ip->ihl * 4, 0);
@@ -396,12 +395,12 @@ vrrp_build_ip(vrrp_rt * vrrp, char *buffer, int buflen)
 
 /* build IPSEC AH header */
 static void
-vrrp_build_ipsecah(vrrp_rt * vrrp, char *buffer, int buflen)
+vrrp_build_ipsecah(vrrp_t * vrrp, char *buffer, int buflen)
 {
 	ICV_mutable_fields *ip_mutable_fields;
 	unsigned char *digest;
 	struct iphdr *ip = (struct iphdr *) (buffer);
-	ipsec_ah *ah = (ipsec_ah *) (buffer + sizeof (struct iphdr));
+	ipsec_ah_t *ah = (ipsec_ah_t *) (buffer + sizeof (struct iphdr));
 
 	/* alloc a temp memory space to stock the ip mutable fields */
 	ip_mutable_fields = (ICV_mutable_fields *) MALLOC(sizeof (ICV_mutable_fields));
@@ -464,7 +463,7 @@ vrrp_build_ipsecah(vrrp_rt * vrrp, char *buffer, int buflen)
 	   => No padding needed.
 	   -- rfc2402.3.3.3.1.1.1 & rfc2401.5
 	 */
-	digest = (unsigned char *) MALLOC(16 * sizeof (unsigned char *));
+	digest = (unsigned char *) MALLOC(16); /*MD5_DIGEST_LENGTH */
 	hmac_md5((unsigned char *) buffer, buflen, vrrp->auth_data, sizeof (vrrp->auth_data)
 		 , digest);
 	memcpy(ah->auth_data, digest, HMAC_MD5_TRUNC);
@@ -480,13 +479,13 @@ vrrp_build_ipsecah(vrrp_rt * vrrp, char *buffer, int buflen)
 
 /* build VRRP header */
 static int
-vrrp_build_vrrp(vrrp_rt * vrrp, int prio, char *buffer)
+vrrp_build_vrrp(vrrp_t * vrrp, int prio, char *buffer)
 {
 	int i = 0;
-	vrrp_pkt *hd = (vrrp_pkt *) buffer;
+	vrrphdr_t *hd = (vrrphdr_t *) buffer;
 	uint32_t *iparr;
 	element e;
-	ip_address *ip_addr;
+	ip_address_t *ip_addr;
 
 	/* Family independant */
 	hd->vers_type = (VRRP_VERSION << 4) | VRRP_PKT_ADVERT;
@@ -519,6 +518,7 @@ vrrp_build_vrrp(vrrp_rt * vrrp, int prio, char *buffer)
 	}
 
 	/* finaly compute vrrp checksum */
+	hd->chksum = 0;
 	hd->chksum = in_csum((u_short *) hd, vrrp_hd_len(vrrp), 0);
 
 	return 0;
@@ -526,9 +526,10 @@ vrrp_build_vrrp(vrrp_rt * vrrp, int prio, char *buffer)
 
 /* build VRRP packet */
 static void
-vrrp_build_pkt(vrrp_rt * vrrp, int prio)
+vrrp_build_pkt(vrrp_t * vrrp, int prio, struct sockaddr_storage *addr)
 {
 	char *bufptr;
+	uint32_t dst;
 	int len;
 
 	/* save reference values */
@@ -537,7 +538,9 @@ vrrp_build_pkt(vrrp_rt * vrrp, int prio)
 
 	if (vrrp->family == AF_INET) {
 		/* build the ip header */
-		vrrp_build_ip(vrrp, VRRP_SEND_BUFFER(vrrp), VRRP_SEND_BUFFER_SIZE(vrrp));
+		dst = (addr) ? inet_sockaddrip4(addr) : 
+			       ((struct sockaddr_in *) &global_data->vrrp_mcast_group4)->sin_addr.s_addr;
+		vrrp_build_ip4(vrrp, bufptr, len, dst);
 
 		/* build the vrrp header */
 		vrrp->send_buffer += vrrp_iphdr_len(vrrp);
@@ -566,12 +569,16 @@ vrrp_build_pkt(vrrp_rt * vrrp, int prio)
 
 /* send VRRP packet */
 static int
-vrrp_send_pkt(vrrp_rt * vrrp)
+vrrp_send_pkt(vrrp_t * vrrp, struct sockaddr_storage *addr)
 {
+	struct sockaddr_storage *src = &vrrp->saddr;
 	struct sockaddr_in6 dst6;
+	struct in6_pktinfo *pkt;
 	struct sockaddr_in dst4;
 	struct msghdr msg;
+	struct cmsghdr *cmsg;
 	struct iovec iov;
+	char cbuf[256];
 
 	/* Build the message data */
 	memset(&msg, 0, sizeof(msg));
@@ -580,32 +587,47 @@ vrrp_send_pkt(vrrp_rt * vrrp)
 	iov.iov_base = VRRP_SEND_BUFFER(vrrp);
 	iov.iov_len = VRRP_SEND_BUFFER_SIZE(vrrp);
 
-	/* Sending path */
-	if (vrrp->family == AF_INET) {
+	/* Unicast sending path */
+	if (addr && addr->ss_family == AF_INET) {
+		msg.msg_name = (struct sockaddr_in *) addr;
+		msg.msg_namelen = sizeof(struct sockaddr_in);
+	} else if (addr && addr->ss_family == AF_INET6) {
+		msg.msg_name = (struct sockaddr_in6 *) addr;
+		msg.msg_namelen = sizeof(struct sockaddr_in6);
+		if (src->ss_family == AF_INET6) {
+			msg.msg_control = cbuf;
+			msg.msg_controllen = CMSG_SPACE(sizeof(struct in6_pktinfo));
+
+			cmsg = CMSG_FIRSTHDR(&msg);
+			cmsg->cmsg_level = IPPROTO_IPV6;
+			cmsg->cmsg_type = IPV6_PKTINFO;
+			cmsg->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
+
+			pkt = (struct in6_pktinfo *) CMSG_DATA(cmsg);
+			memset(pkt, 0, sizeof(struct in6_pktinfo));
+			pkt->ipi6_addr = ((struct sockaddr_in6 *) src)->sin6_addr;
+		}
+	} else if (vrrp->family == AF_INET) { /* Multicast sending path */
 		memset(&dst4, 0, sizeof(dst4));
 		dst4.sin_family = AF_INET;
-		dst4.sin_addr.s_addr = htonl(INADDR_VRRP_GROUP);
-
+		dst4.sin_addr = ((struct sockaddr_in *) &global_data->vrrp_mcast_group4)->sin_addr;
 		msg.msg_name = &dst4;
 		msg.msg_namelen = sizeof(dst4);
 	} else if (vrrp->family == AF_INET6) {
 		memset(&dst6, 0, sizeof(dst6));
 		dst6.sin6_family = AF_INET6;
-		dst6.sin6_port = htons(IPPROTO_VRRP);
-		dst6.sin6_addr.s6_addr16[0] = htons(0xff02);
-		dst6.sin6_addr.s6_addr16[7] = htons(0x12);
-
+		dst6.sin6_addr = ((struct sockaddr_in6 *) &global_data->vrrp_mcast_group6)->sin6_addr;
 		msg.msg_name = &dst6;
 		msg.msg_namelen = sizeof(dst6);
 	}
 
 	/* Send the packet */
-	return sendmsg(vrrp->fd_out, &msg, MSG_DONTROUTE);
+	return sendmsg(vrrp->fd_out, &msg, (addr) ? 0 : MSG_DONTROUTE);
 }
 
 /* Allocate the sending buffer */
 static void
-vrrp_alloc_send_buffer(vrrp_rt * vrrp)
+vrrp_alloc_send_buffer(vrrp_t * vrrp)
 {
 	vrrp->send_buffer_size = vrrp_hd_len(vrrp);
 
@@ -620,8 +642,13 @@ vrrp_alloc_send_buffer(vrrp_rt * vrrp)
 
 /* send VRRP advertissement */
 int
-vrrp_send_adv(vrrp_rt * vrrp, int prio)
+vrrp_send_adv(vrrp_t * vrrp, int prio)
 {
+	struct sockaddr_storage *addr;
+	list l = vrrp->unicast_peer;
+	element e;
+	int ret;
+
 	/* alloc send buffer */
 	if (!vrrp->send_buffer)
 		vrrp_alloc_send_buffer(vrrp);
@@ -629,15 +656,28 @@ vrrp_send_adv(vrrp_rt * vrrp, int prio)
 		memset(vrrp->send_buffer, 0, VRRP_SEND_BUFFER_SIZE(vrrp));
 
 	/* build the packet */
-	vrrp_build_pkt(vrrp, prio);
+	if (!LIST_ISEMPTY(l)) {
+		for (e = LIST_HEAD(l); e; ELEMENT_NEXT(e)) {
+			addr = ELEMENT_DATA(e);
+			vrrp_build_pkt(vrrp, prio, addr);
+			ret = vrrp_send_pkt(vrrp, addr);
+			if (ret < 0) {
+				log_message(LOG_INFO, "VRRP_Instance(%s) Cant sent advert to %s (%m)"
+						    , vrrp->iname, inet_sockaddrtos(addr));
+			}
+		}
+	} else {
+		vrrp_build_pkt(vrrp, prio, NULL);
+		vrrp_send_pkt(vrrp, NULL);
+	}
 
 	/* send it */
-	return vrrp_send_pkt(vrrp);
+	return 0;
 }
 
 /* Received packet processing */
 int
-vrrp_check_packet(vrrp_rt * vrrp, char *buf, int buflen)
+vrrp_check_packet(vrrp_t * vrrp, char *buf, int buflen)
 {
 	int ret;
 
@@ -660,7 +700,7 @@ vrrp_check_packet(vrrp_rt * vrrp, char *buf, int buflen)
 
 /* Gratuitous ARP on each VIP */
 static void
-vrrp_send_update(vrrp_rt * vrrp, ip_address * ipaddress, int idx)
+vrrp_send_update(vrrp_t * vrrp, ip_address_t * ipaddress, int idx)
 {
 	char *msg;
 	char addr_str[41];
@@ -682,10 +722,10 @@ vrrp_send_update(vrrp_rt * vrrp, ip_address * ipaddress, int idx)
 }
 
 void
-vrrp_send_link_update(vrrp_rt * vrrp)
+vrrp_send_link_update(vrrp_t * vrrp)
 {
 	int j;
-	ip_address *ipaddress;
+	ip_address_t *ipaddress;
 	element e;
 
 	/* Only send gratuitous ARP if VIP are set */
@@ -712,7 +752,7 @@ vrrp_send_link_update(vrrp_rt * vrrp)
 
 /* becoming master */
 void
-vrrp_state_become_master(vrrp_rt * vrrp)
+vrrp_state_become_master(vrrp_t * vrrp)
 {
 	/* add the ip addresses */
 	if (!LIST_ISEMPTY(vrrp->vip))
@@ -728,6 +768,11 @@ vrrp_state_become_master(vrrp_rt * vrrp)
 	/* remotes neighbour update */
 	vrrp_send_link_update(vrrp);
 
+	/* set refresh timer */
+	if (vrrp->garp_refresh) {
+		vrrp->garp_refresh_timer = timer_add_long(time_now, vrrp->garp_refresh);
+	}
+
 	/* Check if notify is needed */
 	notify_instance_exec(vrrp, VRRP_STATE_MAST);
 
@@ -742,25 +787,42 @@ vrrp_state_become_master(vrrp_rt * vrrp)
 #endif
 }
 
+/* If the preempt_delay is set we cannot yet transition to master state.  We
+ * must await the timeout of our preempt_delay.  The preemption delay is used
+ * when starting up, or rebooting, a node which needs time to sort out its
+ * routing table (e.g., BGP or OSPF) before it can assume the master role.
+ */
 void
-vrrp_state_goto_master(vrrp_rt * vrrp)
+vrrp_state_goto_master(vrrp_t * vrrp)
 {
+	if (timer_cmp(vrrp->preempt_time, timer_now()) > 0) {
+		vrrp->ms_down_timer = timer_tol(timer_sub(vrrp->preempt_time, timer_now()));
+		return;
+	}
+
 	/*
 	 * Send an advertisement. To force a new master
 	 * election.
 	 */
-        if (vrrp->sync && !vrrp_sync_goto_master(vrrp))
-	      return;
+	if (vrrp->sync && !vrrp_sync_goto_master(vrrp)) {
+		/*
+		 * Set quick sync flag to enable faster transition, i.e. check
+		 * again in the next interval instead of waiting three.
+		 */
+		vrrp->quick_sync = 1;
+		return;
+	}
+
 	vrrp_send_adv(vrrp, vrrp->effective_priority);
 
 	vrrp->state = VRRP_STATE_MAST;
-	log_message(LOG_INFO, "VRRP_Instance(%s) Transition to MASTER STATE",
-	       vrrp->iname);
+	log_message(LOG_INFO, "VRRP_Instance(%s) Transition to MASTER STATE"
+			    , vrrp->iname);
 }
 
 /* leaving master state */
 void
-vrrp_restore_interface(vrrp_rt * vrrp, int advF)
+vrrp_restore_interface(vrrp_t * vrrp, int advF)
 {
         /* if we stop vrrp, warn the other routers to speed up the recovery */
 	if (advF) {
@@ -792,7 +854,7 @@ vrrp_restore_interface(vrrp_rt * vrrp, int advF)
 }
 
 void
-vrrp_state_leave_master(vrrp_rt * vrrp)
+vrrp_state_leave_master(vrrp_t * vrrp)
 {
 	if (VRRP_VIP_ISSET(vrrp)) {
 #ifdef _HAVE_IPVS_SYNCD_
@@ -818,6 +880,7 @@ vrrp_state_leave_master(vrrp_rt * vrrp)
 		vrrp_restore_interface(vrrp, 0);
 		vrrp->state = VRRP_STATE_FAULT;
 		notify_instance_exec(vrrp, VRRP_STATE_FAULT);
+		vrrp_send_adv(vrrp, VRRP_PRIO_STOP);
 #ifdef _WITH_SNMP_
 		vrrp_snmp_instance_trap(vrrp);
 #endif
@@ -830,9 +893,9 @@ vrrp_state_leave_master(vrrp_rt * vrrp)
 
 /* BACKUP state processing */
 void
-vrrp_state_backup(vrrp_rt * vrrp, char *buf, int buflen)
+vrrp_state_backup(vrrp_t * vrrp, char *buf, int buflen)
 {
-	vrrp_pkt *hd;
+	vrrphdr_t *hd;
 	uint32_t saddr;
 	int ret = 0, proto;
 
@@ -859,7 +922,7 @@ vrrp_state_backup(vrrp_rt * vrrp, char *buf, int buflen)
 
 /* MASTER state processing */
 int
-vrrp_state_master_tx(vrrp_rt * vrrp, const int prio)
+vrrp_state_master_tx(vrrp_t * vrrp, const int prio)
 {
 	int ret = 0;
 
@@ -868,6 +931,9 @@ vrrp_state_master_tx(vrrp_rt * vrrp, const int prio)
 				    , vrrp->iname);
 		vrrp_state_become_master(vrrp);
 		ret = 1;
+	} else if (vrrp->garp_refresh && timer_cmp(time_now, vrrp->garp_refresh_timer) > 0) {
+		vrrp_send_link_update(vrrp);
+		vrrp->garp_refresh_timer = timer_add_long(time_now, vrrp->garp_refresh);
 	}
 
 	vrrp_send_adv(vrrp,
@@ -877,12 +943,12 @@ vrrp_state_master_tx(vrrp_rt * vrrp, const int prio)
 }
 
 int
-vrrp_state_master_rx(vrrp_rt * vrrp, char *buf, int buflen)
+vrrp_state_master_rx(vrrp_t * vrrp, char *buf, int buflen)
 {
-	vrrp_pkt *hd = NULL;
+	vrrphdr_t *hd = NULL;
 	int ret = 0, proto = 0;
 	uint32_t saddr = 0;
-	ipsec_ah *ah;
+	ipsec_ah_t *ah;
 
 	/* return on link failure */
 	if (vrrp->wantstate == VRRP_STATE_GOTO_FAULT) {
@@ -907,7 +973,7 @@ vrrp_state_master_rx(vrrp_rt * vrrp, char *buf, int buflen)
 		log_message(LOG_INFO, "VRRP_Instance(%s) Received lower prio advert"
 				      ", forcing new election", vrrp->iname);
 		if (proto == IPPROTO_IPSEC_AH) {
-			ah = (ipsec_ah *) (buf + sizeof(struct iphdr));
+			ah = (ipsec_ah_t *) (buf + sizeof(struct iphdr));
 			log_message(LOG_INFO, "VRRP_Instance(%s) IPSEC-AH : Syncing seq_num"
 					      " - Increment seq"
 					    , vrrp->iname);
@@ -923,11 +989,16 @@ vrrp_state_master_rx(vrrp_rt * vrrp, char *buf, int buflen)
 	} else if (vrrp->family == AF_INET) {
 		if (hd->priority > vrrp->effective_priority ||
 		    (hd->priority == vrrp->effective_priority &&
-		     ntohl(saddr) > VRRP_PKT_SADDR(vrrp))) {
+		     ntohl(saddr) > ntohl(VRRP_PKT_SADDR(vrrp)))) {
+			/* We send a last advert here in order to refresh remote MASTER
+			 * coming up to force link update at MASTER side.
+			 */
+			vrrp_send_adv(vrrp, vrrp->effective_priority);
+
 			log_message(LOG_INFO, "VRRP_Instance(%s) Received higher prio advert"
 					    , vrrp->iname);
 			if (proto == IPPROTO_IPSEC_AH) {
-				ah = (ipsec_ah *) (buf + sizeof(struct iphdr));
+				ah = (ipsec_ah_t *) (buf + sizeof(struct iphdr));
 				log_message(LOG_INFO, "VRRP_Instance(%s) IPSEC-AH : Syncing seq_num"
 						      " - Decrement seq"
 						    , vrrp->iname);
@@ -942,6 +1013,11 @@ vrrp_state_master_rx(vrrp_rt * vrrp, char *buf, int buflen)
 	} else if (vrrp->family == AF_INET6) {
 		/* FIXME: compare v6 saddr to link local when prio are equal !!! */
 		if (hd->priority > vrrp->effective_priority) {
+			/* We send a last advert here in order to refresh remote MASTER
+			 * coming up to force link update at MASTER side.
+			 */
+			vrrp_send_adv(vrrp, vrrp->effective_priority);
+
 			log_message(LOG_INFO, "VRRP_Instance(%s) Received higher prio advert"
 					    , vrrp->iname);
 			vrrp->ms_down_timer = 3 * vrrp->adver_int + VRRP_TIMER_SKEW(vrrp);
@@ -955,9 +1031,9 @@ vrrp_state_master_rx(vrrp_rt * vrrp, char *buf, int buflen)
 }
 
 int
-vrrp_state_fault_rx(vrrp_rt * vrrp, char *buf, int buflen)
+vrrp_state_fault_rx(vrrp_t * vrrp, char *buf, int buflen)
 {
-	vrrp_pkt *hd;
+	vrrphdr_t *hd;
 	uint32_t saddr;
 	int ret = 0, proto;
 
@@ -980,7 +1056,7 @@ vrrp_state_fault_rx(vrrp_rt * vrrp, char *buf, int buflen)
 
 /* check for minimum configuration requirements */
 static int
-chk_min_cfg(vrrp_rt * vrrp)
+chk_min_cfg(vrrp_t * vrrp)
 {
 	if (vrrp->vrid == 0) {
 		log_message(LOG_INFO, "VRRP_Instance(%s) the virtual id must be set!",
@@ -998,12 +1074,12 @@ chk_min_cfg(vrrp_rt * vrrp)
 
 /* open a VRRP sending socket */
 int
-open_vrrp_send_socket(sa_family_t family, int proto, int idx)
+open_vrrp_send_socket(sa_family_t family, int proto, int idx, int unicast)
 {
-	interface *ifp;
+	interface_t *ifp;
 	int fd = -1;
 
-	/* Retreive interface */
+	/* Retreive interface_t */
 	ifp = if_get_by_ifindex(idx);
 
 	/* Create and init socket descriptor */
@@ -1017,15 +1093,18 @@ open_vrrp_send_socket(sa_family_t family, int proto, int idx)
 		/* Set v4 related */
 		if_setsockopt_hdrincl(&fd);
 		if_setsockopt_bindtodevice(&fd, ifp);
-		if_setsockopt_mcast_loop(family, &fd);
+		if (!unicast)
+			if_setsockopt_mcast_loop(family, &fd);
 		if_setsockopt_priority(&fd);
 		if (fd < 0)
 			return -1;
 	} else if (family == AF_INET6) {
 		/* Set v6 related */
-		if_setsockopt_mcast_hops(family, &fd);
-		if_setsockopt_mcast_if(family, &fd, ifp);
-		if_setsockopt_mcast_loop(family, &fd);
+		if (!unicast) {
+			if_setsockopt_mcast_hops(family, &fd);
+			if_setsockopt_mcast_if(family, &fd, ifp);
+			if_setsockopt_mcast_loop(family, &fd);
+		}
 		if_setsockopt_priority(&fd);
 		if (fd < 0)
 			return -1;
@@ -1041,12 +1120,12 @@ open_vrrp_send_socket(sa_family_t family, int proto, int idx)
 
 /* open a VRRP socket and join the multicast group. */
 int
-open_vrrp_socket(sa_family_t family, int proto, int idx)
+open_vrrp_socket(sa_family_t family, int proto, int idx, int unicast)
 {
-	interface *ifp;
+	interface_t *ifp;
 	int fd = -1;
 
-	/* Retreive interface */
+	/* Retreive interface_t */
 	ifp = if_get_by_ifindex(idx);
 
 	/* open the socket */
@@ -1058,7 +1137,8 @@ open_vrrp_socket(sa_family_t family, int proto, int idx)
 	}
 
 	/* Join the VRRP MCAST group */
-	if_join_vrrp_group(family, &fd, ifp, proto);
+	if (!unicast)
+		if_join_vrrp_group(family, &fd, ifp, proto);
 	if (fd < 0)
 		return -1;
 
@@ -1071,24 +1151,31 @@ open_vrrp_socket(sa_family_t family, int proto, int idx)
 }
 
 void
-close_vrrp_socket(vrrp_rt * vrrp)
+close_vrrp_socket(vrrp_t * vrrp)
 {
-	if_leave_vrrp_group(vrrp->family, vrrp->fd_in, vrrp->ifp);
+	if (LIST_ISEMPTY(vrrp->unicast_peer)) {
+		if_leave_vrrp_group(vrrp->family, vrrp->fd_in, vrrp->ifp);
+	} else {
+		close(vrrp->fd_in);
+	}
 	close(vrrp->fd_out);
 }
 
 int
-new_vrrp_socket(vrrp_rt * vrrp)
+new_vrrp_socket(vrrp_t * vrrp)
 {
 	int old_fd = vrrp->fd_in;
-	int proto;
+	int proto, ifindex, unicast;
 
 	/* close the desc & open a new one */
 	close_vrrp_socket(vrrp);
 	remove_vrrp_fd_bucket(vrrp);
 	proto = (vrrp->auth_type == VRRP_AUTH_AH) ? IPPROTO_IPSEC_AH : IPPROTO_VRRP;
-	vrrp->fd_in = open_vrrp_socket(vrrp->family, proto, IF_INDEX(vrrp->ifp));
-	vrrp->fd_out = open_vrrp_send_socket(vrrp->family, proto, IF_INDEX(vrrp->ifp));
+	ifindex = (vrrp->vmac_flags & VRRP_VMAC_FL_XMITBASE) ? IF_BASE_INDEX(vrrp->ifp) :
+							       IF_INDEX(vrrp->ifp);
+	unicast = !LIST_ISEMPTY(vrrp->unicast_peer);
+	vrrp->fd_in = open_vrrp_socket(vrrp->family, proto, ifindex, unicast);
+	vrrp->fd_out = open_vrrp_send_socket(vrrp->family, proto, ifindex, unicast);
 	alloc_vrrp_fd_bucket(vrrp);
 
 	/* Sync the other desc */
@@ -1103,7 +1190,7 @@ shutdown_vrrp_instances(void)
 {
 	list l = vrrp_data->vrrp;
 	element e;
-	vrrp_rt *vrrp;
+	vrrp_t *vrrp;
 
 	for (e = LIST_HEAD(l); e; ELEMENT_NEXT(e)) {
 		vrrp = ELEMENT_DATA(e);
@@ -1113,7 +1200,7 @@ shutdown_vrrp_instances(void)
 			vrrp_restore_interface(vrrp, 1);
 
 		/* Remove VMAC */
-		if (vrrp->vmac)
+		if (vrrp->vmac_flags & VRRP_VMAC_FL_SET)
 			netlink_link_del_vmac(vrrp);
 
 		/* Run stop script */
@@ -1138,7 +1225,7 @@ shutdown_vrrp_instances(void)
 
 /* complete vrrp structure */
 static int
-vrrp_complete_instance(vrrp_rt * vrrp)
+vrrp_complete_instance(vrrp_t * vrrp)
 {
 	vrrp->state = VRRP_STATE_INIT;
 	if (!vrrp->adver_int)
@@ -1154,8 +1241,8 @@ vrrp_complete_init(void)
 {
 	list l;
 	element e;
-	vrrp_rt *vrrp;
-	vrrp_sgroup *sgroup;
+	vrrp_t *vrrp;
+	vrrp_sgroup_t *sgroup;
 
 	/* Complete VRRP instance initialization */
 	l = vrrp_data->vrrp;
@@ -1175,13 +1262,32 @@ vrrp_complete_init(void)
 	return 1;
 }
 
+int
+vrrp_ipvs_needed(void)
+{
+	vrrp_t *vrrp;
+	element e;
+
+	if (!vrrp_data)
+		return 0;
+
+	for (e = LIST_HEAD(vrrp_data->vrrp); e; ELEMENT_NEXT(e)) {
+		vrrp = ELEMENT_DATA(e);
+		if (vrrp->lvs_syncd_if) {
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 /* Try to find a VRRP instance */
-static vrrp_rt *
-vrrp_exist(vrrp_rt * old_vrrp)
+static vrrp_t *
+vrrp_exist(vrrp_t * old_vrrp)
 {
 	element e;
 	list l = vrrp_data->vrrp;
-	vrrp_rt *vrrp;
+	vrrp_t *vrrp;
 
 	if (LIST_ISEMPTY(l))
 		return NULL;
@@ -1197,9 +1303,9 @@ vrrp_exist(vrrp_rt * old_vrrp)
 
 /* Clear VIP|EVIP not present into the new data */
 static void
-clear_diff_vrrp_vip(vrrp_rt * old_vrrp, int type)
+clear_diff_vrrp_vip(vrrp_t * old_vrrp, int type)
 {
-	vrrp_rt *vrrp = vrrp_exist(old_vrrp);
+	vrrp_t *vrrp = vrrp_exist(old_vrrp);
 	list l = (type == VRRP_VIP_TYPE) ? old_vrrp->vip : old_vrrp->evip;
 	list n = (type == VRRP_VIP_TYPE) ? vrrp->vip : vrrp->evip;
 	clear_diff_address(l, n);
@@ -1207,24 +1313,24 @@ clear_diff_vrrp_vip(vrrp_rt * old_vrrp, int type)
 
 /* Clear virtual routes not present in the new data */
 static void
-clear_diff_vrrp_vroutes(vrrp_rt * old_vrrp)
+clear_diff_vrrp_vroutes(vrrp_t * old_vrrp)
 {
-	vrrp_rt *vrrp = vrrp_exist(old_vrrp);
+	vrrp_t *vrrp = vrrp_exist(old_vrrp);
 	clear_diff_routes(old_vrrp->vroutes, vrrp->vroutes);
 }
 
 /* Keep the state from before reload */
 static void
-reset_vrrp_state(vrrp_rt * old_vrrp)
+reset_vrrp_state(vrrp_t * old_vrrp)
 {
 	/* Keep VRRP state, ipsec AH seq_number */
-	vrrp_rt *vrrp = vrrp_exist(old_vrrp);
+	vrrp_t *vrrp = vrrp_exist(old_vrrp);
 	vrrp->state = old_vrrp->state;
 	vrrp->init_state = old_vrrp->state;
 	vrrp->wantstate = old_vrrp->state;
 	if (!old_vrrp->sync)
 		vrrp->effective_priority = old_vrrp->effective_priority;
-	memcpy(vrrp->ipsecah_counter, old_vrrp->ipsecah_counter, sizeof(seq_counter));
+	memcpy(vrrp->ipsecah_counter, old_vrrp->ipsecah_counter, sizeof(seq_counter_t));
 
 #ifdef _HAVE_IPVS_SYNCD_
 	if (old_vrrp->lvs_syncd_if)
@@ -1257,14 +1363,14 @@ clear_diff_vrrp(void)
 {
 	element e;
 	list l = old_vrrp_data->vrrp;
-	vrrp_rt *vrrp;
+	vrrp_t *vrrp;
 
 	if (LIST_ISEMPTY(l))
 		return;
 
 	for (e = LIST_HEAD(l); e; ELEMENT_NEXT(e)) {
 		vrrp = ELEMENT_DATA(e);
-		vrrp_rt *new_vrrp;
+		vrrp_t *new_vrrp;
 
 		/*
 		 * Try to find this vrrp into the new conf data
@@ -1275,7 +1381,7 @@ clear_diff_vrrp(void)
 			vrrp_restore_interface(vrrp, 1);
 
 			/* Remove VMAC if one was created */
-			if (vrrp->vmac) 
+			if (vrrp->vmac_flags & VRRP_VMAC_FL_SET) 
 				netlink_link_del_vmac(vrrp);
 		} else {
 			/*
@@ -1292,7 +1398,8 @@ clear_diff_vrrp(void)
 			 * Remove VMAC if it existed in old vrrp instance,
 			 * but not the new one.
 			 */
-			if (vrrp->vmac && !new_vrrp->vmac) {
+			if (vrrp->vmac_flags & VRRP_VMAC_FL_SET &&
+			    !(new_vrrp->vmac_flags & VRRP_VMAC_FL_SET)) {
 				netlink_link_del_vmac(vrrp);
 			}
 
@@ -1308,7 +1415,7 @@ clear_diff_script(void)
 {
 	element e;
 	list l = old_vrrp_data->vrrp_script;
-	vrrp_script *vscript, *nvscript;
+	vrrp_script_t *vscript, *nvscript;
 
 	if (LIST_ISEMPTY(l))
 		return;
